/**
 * Simulation of process scheduling algorithms
 *
 * Ron Mak
 * Department of Computer Engineering
 * San Jose State University
 */
#include <iostream>
#include <algorithm>
#include "SRT.h"

using namespace std;

void SRT::prepare_run(int run_index)
{
    cout << endl;
    cout << "-----------------------------------" << endl;
    cout << "Shortest Remaining Time First (SRT): run #" << run_index << endl;
    cout << "-----------------------------------" << endl;
}

void SRT::run_algorithm()
{
    int quantum = 0;

    // Nothing to do until the first process arrives.
    while (quantum < job_queue[0]->arrival_time)
    {
        timeline += '-';
        quantum++;
    }

    vector<Process *> new_arrivals;

    // Run all the processes up to the quantum limit.
    do
    {
        new_arrivals.clear();

        // Keep track of new arrivals during this quantum.
        while (   (job_queue.size() > 0)
               && (quantum >= job_queue.front()->arrival_time))
        {
            Process *p = job_queue.front();
            p->stats.wait_time += quantum - p->arrival_time;
            p->start_time = quantum;
            p->previous_quantum = quantum - 1;

            new_arrivals.push_back(p);
            job_queue.erase(job_queue.begin());
        }

        // Sort any new arrivals by expected run time.
        if (new_arrivals.size() > 1)
        {
            sort(new_arrivals.begin(), new_arrivals.end(),
                 [] (Process *p1, Process *p2)
                    { return p1->expected_run_time < p2->expected_run_time; });
        }

        if (ready_queue.size() == 0)
        {
            // Move any new arrivals into the ready queue.
            if (new_arrivals.size() == 0) timeline += '-';
            else                          ready_queue = move(new_arrivals);
        }
        else if (new_arrivals.size() > 0)
        {
            // If a new arrival's expected run time is less than
            // the remaining run time of the currently running process,
            // preempt the currently running process. Otherwise, append
            // the new arrivals to the end of the ready queue.
            if (new_arrivals.front()->expected_run_time
                                        < ready_queue.front()->remaining_time)
            {
                // Preempt the currently running process.
                ready_queue.front()->previous_quantum = quantum;
                ready_queue.push_back(ready_queue.front());
                ready_queue.front() = new_arrivals.front();
                new_arrivals.erase(new_arrivals.begin());
            }

            // Append remaining new arrivals to the end of the ready queue.
            while (new_arrivals.size() > 0)
            {
                ready_queue.push_back(new_arrivals.front());
                new_arrivals.erase(new_arrivals.begin());
            }
        }

        // Run the process at the head of the ready queue.
        if (ready_queue.size() > 0)
        {
            Process *p = ready_queue.front();
            p->stats.wait_time += quantum - p->previous_quantum - 1;
            p->previous_quantum = quantum;
            timeline += p->name;

            // Process finished?
            if (p->remaining_time <= 1.0)
            {
                // Yes: Remove from the ready queue.
                ready_queue.erase(ready_queue.begin());
                completed.push_back(p);
                p->stats.turnaround_time = quantum + p->remaining_time
                                                            - p->arrival_time;
                end_time = quantum + p->remaining_time;
            }
            else
            {
                // No
                p->remaining_time -= 1;
            }
        }
    } while (++quantum < quantum_limit);
}
